# Generated by Django 3.2.22 on 2024-01-13 20:08

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("NEMO_publications", "0002_publicationmetadata_month"),
    ]

    operations = [
        migrations.AddField(
            model_name="publicationmetadata",
            name="decision_time",
            field=models.DateTimeField(
                blank=True,
                help_text="The date and time when the publication metadata was approved or rejected.",
                null=True,
            ),
        ),
        migrations.AddField(
            model_name="publicationmetadata",
            name="json_metadata",
            field=models.TextField(blank=True, help_text="Publication json metadata.", null=True),
        ),
        migrations.AddField(
            model_name="publicationmetadata",
            name="status",
            field=models.IntegerField(
                blank=True,
                choices=[(1, "Approved"), (2, "Rejected")],
                help_text="Publication metadata status.",
                null=True,
            ),
        ),
    ]



from django.contrib import admin
from django.contrib.admin import register
from django.utils import timezone

from NEMO_publications.forms import PublicationDataForm, UserPublicationStatusForm
from NEMO_publications.models import (
    PublicationData,
    PublicationMetadata,
    PublicationMetadataStatus,
    UserPublicationStatus,
)
from NEMO_publications.utils import fetch_publication_metadata_by_doi
from NEMO_publications.views.publications import export_publications


@admin.action(description="Export selected publications in CSV")
def export_publication_for_metadata(model_admin, request, queryset):
    return export_publications(queryset.all())


@admin.action(description="Reject selected publications")
def reject_publication_metadata(model_admin, request, queryset):
    queryset.all().update(status=PublicationMetadataStatus.REJECTED, decision_time=timezone.now())


@admin.action(description="Approve selected publications")
def approve_publication_metadata(model_admin, request, queryset):
    queryset.all().update(status=PublicationMetadataStatus.APPROVED, decision_time=timezone.now())


@admin.action(description="Delete all associated publication data")
def delete_publication_data(model_admin, request, queryset):
    PublicationData.objects.filter(metadata__in=queryset.all()).delete()


@admin.action(description="Delete all publication metadata without data entries")
def delete_metadata_without_data(model_admin, request, queryset):
    PublicationMetadata.objects.filter(publicationdata=None).delete()


@admin.action(
    description="Fetch metadata for selected publications from remote source using DOI (Replace existing field values with remote values)"
)
def fetch_and_merge_metadata(model_admin, request, queryset):
    for metadata in queryset.all():
        fetch_result = fetch_publication_metadata_by_doi(metadata.doi)
        if "error" not in fetch_result.keys():
            cleaned_fetched_metadata = {k: v for k, v in fetch_result["metadata"].items() if v}
            del cleaned_fetched_metadata["doi"]
            for key, value in cleaned_fetched_metadata.items():
                setattr(metadata, key, value)
            try:
                metadata.save()
            except Exception:
                pass


class StatusFilter(admin.SimpleListFilter):
    title = "Status"
    parameter_name = "status"

    def lookups(self, request, model_admin):
        return [
            ("none", "None"),
            *PublicationMetadataStatus.Choices,
        ]

    def queryset(self, request, queryset):
        if not self.value():
            return queryset
        elif self.value() == "none":
            return queryset.filter(status__isnull=True)
        else:
            return queryset.filter(status=self.value())


@register(PublicationMetadata)
class PublicationMetadataAdmin(admin.ModelAdmin):
    list_display = (
        "title",
        "journal",
        "month",
        "year",
        "doi",
    )
    list_filter = (
        StatusFilter,
        "year",
        "month",
        ("creator", admin.RelatedOnlyFieldListFilter),
    )
    date_hierarchy = "creation_time"
    actions = [
        export_publication_for_metadata,
        delete_publication_data,
        delete_metadata_without_data,
        reject_publication_metadata,
        approve_publication_metadata,
        fetch_and_merge_metadata,
    ]
    readonly_fields = ["creation_time"]


@register(PublicationData)
class PublicationDataAdmin(admin.ModelAdmin):
    list_display = (
        "creator",
        "creation_time",
        "metadata",
        "get_authors",
        "get_tools",
        "get_projects",
    )
    filter_horizontal = (
        "authors",
        "tools",
        "projects",
    )
    form = PublicationDataForm
    list_filter = [("creator", admin.RelatedOnlyFieldListFilter), ("metadata", admin.RelatedOnlyFieldListFilter)]
    date_hierarchy = "creation_time"

    def get_authors(self, data):
        return data.get_authors()

    def get_tools(self, data):
        return data.get_tools()

    def get_projects(self, data):
        return data.get_projects()


@register(UserPublicationStatus)
class UserPublicationStatusAdmin(admin.ModelAdmin):
    list_display = ["user", "doi", "date_added", "status"]
    form = UserPublicationStatusForm
    list_filter = ["date_added", "status"]
    date_hierarchy = "date_added"
    search_fields = ["doi"]



# Generated by Django 3.2.22 on 2023-11-15 02:17

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ("NEMO", "0054_documents_update_max_length"),
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name="PublicationMetadata",
            fields=[
                ("id", models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                (
                    "creation_time",
                    models.DateTimeField(
                        auto_now_add=True, help_text="The date and time when the publication metadata was created."
                    ),
                ),
                (
                    "doi",
                    models.CharField(
                        help_text="Digital Object Identifier (DOI) of the publication.", max_length=255, unique=True
                    ),
                ),
                ("title", models.CharField(help_text="Title of the publication.", max_length=255)),
                ("journal", models.CharField(help_text="Journal where the publication is published.", max_length=255)),
                ("bibtex", models.TextField(blank=True, help_text="Publication bibtex metadata.", null=True)),
                ("year", models.PositiveIntegerField(help_text="Publication year.")),
                (
                    "creator",
                    models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to=settings.AUTH_USER_MODEL),
                ),
            ],
            options={
                "abstract": False,
            },
        ),
        migrations.CreateModel(
            name="PublicationData",
            fields=[
                ("id", models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                (
                    "creation_time",
                    models.DateTimeField(
                        auto_now_add=True, help_text="The date and time when the publication data was created."
                    ),
                ),
                (
                    "authors",
                    models.ManyToManyField(
                        blank=True, help_text="Authors of the publication.", to=settings.AUTH_USER_MODEL
                    ),
                ),
                (
                    "creator",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="publication_data_creator",
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
                (
                    "metadata",
                    models.ForeignKey(
                        help_text="The metadata of the publication",
                        on_delete=django.db.models.deletion.CASCADE,
                        to="NEMO_publications.publicationmetadata",
                    ),
                ),
                (
                    "projects",
                    models.ManyToManyField(
                        blank=True, help_text="Projects involved in the publication.", to="NEMO.Project"
                    ),
                ),
                (
                    "tools",
                    models.ManyToManyField(blank=True, help_text="Tools involved in the publication.", to="NEMO.Tool"),
                ),
            ],
            options={
                "abstract": False,
            },
        ),
    ]



import json
import math
from typing import Optional
from urllib.parse import urlparse

from NEMO.decorators import staff_member_required
from NEMO.models import Project, Tool, User
from NEMO.typing import QuerySetType
from NEMO.utilities import (
    BasicDisplayTable,
    export_format_datetime,
    queryset_search_filter,
    quiet_int,
    render_combine_responses,
)
from NEMO.views.landing import landing
from NEMO.views.notifications import get_notifications
from NEMO.views.pagination import SortedPaginator
from django.contrib.auth.decorators import login_required
from django.db.models import Count, Q
from django.http import HttpResponse, HttpResponseBadRequest, HttpResponseNotFound
from django.shortcuts import get_object_or_404, redirect, render
from django.urls import reverse
from django.utils import timezone
from django.utils.safestring import mark_safe
from django.views.decorators.http import require_GET, require_POST, require_http_methods

from NEMO_publications.customization import PublicationsCustomization
from NEMO_publications.forms import (
    PublicationDataForm,
    create_publication_data_post_form,
    create_publication_metadata_post_form,
)
from NEMO_publications.models import PublicationData, PublicationMetadata, PublicationMetadataStatus
from NEMO_publications.utils import (
    PUBLICATION_NOTIFICATION,
    fetch_publication_metadata_by_doi,
)
from NEMO_publications.models import UserPublicationStatus


@login_required
@require_GET
def new_landing(request):
    original_view = landing(request)
    if PublicationsCustomization.get_bool("publications_enable_landing_carousel") or PublicationsCustomization.get_bool(
        "publications_enable_landing_warning_pending_publications"
    ):
        return render_combine_responses(
            request,
            original_view,
            "NEMO_publications/new_landing.html",
            {"publication_notifications": get_notifications(request.user, PUBLICATION_NOTIFICATION, delete=False)},
        )
    else:
        return original_view


@login_required
@require_GET
def user_search(request):
    return queryset_search_filter(User.objects.all(), ["first_name", "last_name", "username"], request)


@staff_member_required
@require_POST
def set_publication_status(request, publication_metadata_id):
    publication_metadata = get_object_or_404(PublicationMetadata, id=publication_metadata_id)
    decision_option = [state for state in ["approve", "reject"] if state in request.POST]
    decision = decision_option[0] if len(decision_option) == 1 else None
    if publication_metadata.status:
        return HttpResponseBadRequest("Publication metadata status is already set")
    if decision and not publication_metadata.status:
        if decision == "approve":
            publication_metadata.status = PublicationMetadataStatus.APPROVED
        elif decision == "reject":
            publication_metadata.status = PublicationMetadataStatus.REJECTED
        publication_metadata.decision_time = timezone.now()
        publication_metadata.save()
    return HttpResponse()


@login_required
@require_http_methods(["GET", "POST"])
def create_or_update_publication_data(request, publication_data_id=None, publication_metadata_id=None):
    user: User = request.user
    dictionary = {}

    # Initialize publication data (found: edit, not found: create) and publication metadata (mandatory)
    try:
        publication_data = PublicationData.objects.get(id=publication_data_id)
        publication_metadata = PublicationMetadata.objects.get(id=publication_data.metadata.id)
    except PublicationData.DoesNotExist:
        publication_data = PublicationData()
        # Initialize publication metadata from POST["metadata_id"] - when creating a new publication data entry
        if publication_metadata_id:
            publication_metadata = PublicationMetadata.objects.get(id=publication_metadata_id)
        else:
            return HttpResponseBadRequest("Publication metadata not found.")
    except PublicationMetadata.DoesNotExist:
        return HttpResponseBadRequest("Publication metadata is invalid.")

    if publication_metadata.status == PublicationMetadataStatus.REJECTED:
        return HttpResponseBadRequest("You cannot edit this publication.")

    if request.method == "POST":
        edit = bool(publication_data.id)
        # Only creator can edit publication_data
        if edit and publication_data.creator != user:
            return HttpResponseNotFound("Publication data not found.")
        form = create_publication_data_post_form(request, user, edit, publication_data)
        # Initialize creator and metadata - when creating a new publication data entry
        if not edit:
            form.instance.creator = user
            form.instance.metadata = publication_metadata

        # Only save if the user is actually adding new info. Unless we don't already have at least on publication
        if (
            not "authors" in form.data and not "projects" in form.data and not "tools" in form.data
        ) and publication_metadata.publicationdata_set.exclude(id=publication_data.id).exists():
            return redirect("publications")
        if form.is_valid():
            form.save()
            if (
                not edit
                and UserPublicationStatus.objects.filter(
                    user=user, status=UserPublicationStatus.Status.PENDING
                ).exists()
            ):
                return redirect("search_publication")
            return redirect("publications")
        else:
            dictionary["form"] = form
            return edit_publication_data(
                request, publication_metadata, dictionary, user, [publication_data.id] if edit else None
            )
    else:
        form = PublicationDataForm(instance=publication_data)
        dictionary["form"] = form
        return edit_publication_data(
            request, publication_metadata, dictionary, user, [publication_data.id] if publication_data.id else None
        )


@login_required
@require_POST
def create_publication_metadata(request):
    user: User = request.user
    metadata_form = create_publication_metadata_post_form(request.POST, user, PublicationMetadata())
    # Can only create publication metadata once, prevent if DOI already exists
    doi = request.POST.get("doi")
    if doi and PublicationMetadata.objects.filter(doi__iexact=doi).exists():
        metadata_form.add_error(None, ["Publication metadata already exists."])
    metadata_form.instance.creator = user
    metadata_form.instance.bibtex = None
    if metadata_form.is_valid():
        publication_metadata = metadata_form.save()
        # Proceed to show publication data form once metadata saved
        return edit_publication_data(
            request,
            publication_metadata,
            {},
            user,
        )
    else:
        return render(request, "NEMO_publications/publication_metadata.html", {"form": metadata_form})


@login_required
@require_http_methods(["GET", "POST"])
def search_publication_by_doi(request):
    user: User = request.user
    user_publications = UserPublicationStatus.objects.filter(user=user, status=UserPublicationStatus.Status.PENDING)
    if request.method == "POST":
        # DOI is required for search
        doi = request.POST.get("doi")
        if not doi:
            return render(request, "NEMO_publications/publication_search.html", {"error": "DOI is missing"})
        # Remove scheme from doi URL if provided, so we only keep the actual DOI
        parsed_url = urlparse(doi)
        doi = parsed_url.geturl().replace("%s://doi.org/" % parsed_url.scheme, "", 1)
        try:
            # Lookup publication metadata first in database using DOI
            publication_metadata = PublicationMetadata.objects.get(doi__iexact=doi)
            # Do not allow to add rejected publication
            if publication_metadata.status == PublicationMetadataStatus.REJECTED:
                return render(
                    request,
                    "NEMO_publications/publication_search.html",
                    {"error": "You cannot add this publication."},
                )
            try:
                # If the user already has a publication data entry for DOI, redirect to publication data edit
                existing_publication_data = PublicationData.objects.get(
                    metadata__id=publication_metadata.id, creator=user
                )
                return redirect("edit_publication", existing_publication_data.id)
            except PublicationData.DoesNotExist:
                return edit_publication_data(
                    request,
                    publication_metadata,
                    {"form": PublicationDataForm(instance=PublicationData())},
                    user,
                )
        except PublicationMetadata.DoesNotExist:
            # If metadata for DOI does not exist in DB, fetch from remote
            publication_metadata_form = fetch_publication_metadata_as_form(doi, user)
            # Always show form so the user can set the image URL if they want
            return render(
                request,
                "NEMO_publications/publication_metadata.html",
                {
                    "form": publication_metadata_form,
                    "user_publication": get_user_publication(user, doi),
                },
            )
    else:
        return render(request, "NEMO_publications/publication_search.html", {"user_publications": user_publications})


def fetch_publication_metadata_as_form(doi, user):
    publication_metadata_search = fetch_publication_metadata_by_doi(doi)
    publication_metadata_form = create_publication_metadata_post_form(
        publication_metadata_search["metadata"], user, PublicationMetadata()
    )
    publication_metadata_form.instance.creator = user
    if "error" in publication_metadata_search.keys():
        publication_metadata_form.add_error(None, [publication_metadata_search["error"]])
    return publication_metadata_form


@login_required
@require_GET
def get_publication_leaderboard(request, year=None):
    current_year = timezone.now().year
    selected_year = quiet_int(year, current_year)
    authors_publication_count = User.objects.annotate(
        publication_count=Count(
            "publicationdata__metadata",
            filter=Q(publicationdata__metadata__year=selected_year)
            & ~Q(publicationdata__metadata__status=PublicationMetadataStatus.REJECTED),
        )
    ).filter(publication_count__gt=0)

    page = SortedPaginator(authors_publication_count, request, order_by="-publication_count").get_current_page()
    years = list(PublicationData.objects.values_list("metadata__year", flat=True).distinct())

    if current_year not in years:
        years.insert(0, current_year)
    years.sort(reverse=True)

    return render(
        request,
        "NEMO_publications/leaderboard.html",
        {
            "page": page,
            "selected_year": selected_year,
            "years": years,
            "publication_notifications": get_notifications(request.user, PUBLICATION_NOTIFICATION, delete=False),
        },
    )


@login_required
@require_GET
def get_publications(request):
    user: User = request.user
    # Get all publications that have data entries
    publications_with_data = (
        PublicationMetadata.objects.exclude(publicationdata=None)
        .exclude(status=PublicationMetadataStatus.REJECTED)
        .prefetch_related(
            "publicationdata_set",
            "publicationdata_set__authors",
            "publicationdata_set__projects",
            "publicationdata_set__tools",
        )
    )

    selected_tool = Tool.objects.filter(id=request.GET.get("selected_tool") or None).first()
    selected_project = Project.objects.filter(id=request.GET.get("selected_project") or None).first()
    selected_author = User.objects.filter(id=request.GET.get("selected_author") or None).first()

    if selected_author:
        publications_with_data = publications_with_data.filter(publicationdata__authors__in=[selected_author])
    if selected_tool:
        publications_with_data = publications_with_data.filter(publicationdata__tools__in=[selected_tool])
    if selected_project:
        publications_with_data = publications_with_data.filter(publicationdata__projects__in=[selected_project])

    page = SortedPaginator(publications_with_data, request, order_by="-creation_time").get_current_page()

    if bool(request.GET.get("csv", False)) and user.is_any_part_of_staff:
        return export_publications(publications_with_data.order_by("-creation_time"))

    # Create dictionary associating each publication metadata id with the user's data entry
    user_publication_data = PublicationData.objects.filter(creator=user, metadata__in=page)
    user_publication_data_dict = {}
    for user_publication_data in user_publication_data:
        user_publication_data_dict[user_publication_data.metadata_id] = user_publication_data

    return render(
        request,
        "NEMO_publications/publications.html",
        {
            "page": page,
            "user_data": user_publication_data_dict,
            "projects": get_json_project_search_list(user),
            "tools": Tool.objects.all(),
            "selected_tool": selected_tool,
            "selected_author": selected_author,
            "selected_project": selected_project,
            "publication_notifications": get_notifications(request.user, PUBLICATION_NOTIFICATION, delete=False),
        },
    )


@login_required
@require_POST
def delete_publication(request, publication_data_id):
    user = request.user
    # Lookup and delete the user's data entry
    publication = get_object_or_404(PublicationData, pk=publication_data_id)
    if publication.creator != user:
        return HttpResponseBadRequest("You are not allowed to delete this publication data")
    publication.delete()
    return redirect("publications")


def edit_publication_data(request, metadata, dictionary, user, other_users_exclude_data=None):
    dictionary["metadata"] = metadata
    dictionary["projects"] = get_json_project_search_list(user)
    dictionary["tools"] = Tool.objects.all()
    dictionary["authors_suggestion"] = metadata.get_bibtex_authors()
    dictionary["other_users"] = metadata.get_related_data(other_users_exclude_data)
    dictionary["user_publication"] = get_user_publication(user, metadata.doi)
    return render(request, "NEMO_publications/publication_data.html", dictionary)


def get_json_project_search_list(user):
    projects = Project.objects.all().prefetch_related("manager_set")
    if not user.is_any_part_of_staff:
        projects = Project.objects.filter(Q(user=user) | Q(manager_set=user))

    search_list = []
    for project in projects:
        search_list.append(
            {
                "id": project.id,
                "name": project.__str__(),
                "pis": [{"id": pi.id, "name": pi.__str__()} for pi in project.manager_set.all()],
            }
        )
    return mark_safe(json.dumps(search_list))


def export_publications(publication_list: QuerySetType[PublicationMetadata]):
    table = get_publications_table_display(publication_list)
    filename = f"publications_{export_format_datetime()}.csv"
    response = table.to_csv()
    response["Content-Disposition"] = f'attachment; filename="{filename}"'
    return response


def get_publications_table_display(publication_list: QuerySetType[PublicationMetadata]) -> BasicDisplayTable:
    table = BasicDisplayTable()
    table.add_header(("title", "Title"))
    table.add_header(("journal", "Journal"))
    table.add_header(("year", "Year"))
    table.add_header(("doi", "DOI"))
    table.add_header(("authors", "Authors"))
    table.add_header(("tools", "Tools"))
    table.add_header(("projects", "Projects"))
    table.add_header(("metadata_creators", "Metadata Created By"))
    table.add_header(("data_creators", "Data Created By"))
    table.add_header(("bibtex", "Bibtex"))
    for metadata in publication_list:
        row = {
            "title": metadata.title,
            "journal": metadata.journal,
            "year": metadata.year,
            "doi": metadata.doi,
            "authors": ", ".join([author.__str__() for author in metadata.get_authors()]),
            "tools": ", ".join([tool.__str__() for tool in metadata.get_tools()]),
            "projects": ", ".join([project.__str__() for project in metadata.get_projects()]),
            "metadata_creators": metadata.creator.__str__(),
            "data_creators": ", ".join([project.__str__() for project in metadata.get_data_creators()]),
            "bibtex": metadata.bibtex,
        }
        table.add_row(row)
    return table


@login_required
@require_GET
def carousel_content(request, content: str = "mixed"):
    selected_year = timezone.now().year
    # Find the first 2 years (going backwards from today) with publications (check 5 years total)
    publication_counts = []
    for year in range(selected_year, selected_year - 5, -1):
        publication_count = (
            User.objects.annotate(
                publication_count=Count(
                    "publicationdata__metadata",
                    filter=Q(publicationdata__metadata__year=year)
                    & ~Q(publicationdata__metadata__status=PublicationMetadataStatus.REJECTED),
                )
            )
            .order_by("-publication_count")
            .filter(publication_count__gt=0)[:5]
        )
        if publication_count:
            publication_counts.append((year, publication_count))
        if len(publication_counts) == 2:
            break
    number_of_latest_pub_to_display = PublicationsCustomization.get_int("publications_carousel_latest_display_count")
    publication_order = json.loads(PublicationsCustomization.get("publications_carousel_latest_display_order"))
    latest_publications = (
        PublicationMetadata.objects.exclude(publicationdata__isnull=True)
        .prefetch_related(
            "publicationdata_set",
            "publicationdata_set__authors",
            "publicationdata_set__projects",
            "publicationdata_set__tools",
        )
        .exclude(status=PublicationMetadataStatus.REJECTED)
        .order_by(*publication_order)[:number_of_latest_pub_to_display]
    )
    is_coming_from_landing = request.META.get("HTTP_REFERER") == request.build_absolute_uri(reverse("landing"))
    return render(
        request,
        "NEMO_publications/carousel.html",
        {
            "publication_counts": publication_counts,
            "publications": latest_publications,
            "show_pub_date": PublicationsCustomization.get_bool("publications_carousel_show_date"),
            "show_pub_authors": PublicationsCustomization.get_bool("publications_carousel_show_authors"),
            "show_pub_tools": PublicationsCustomization.get_bool("publications_carousel_show_tools"),
            "show_pub_projects": PublicationsCustomization.get_bool("publications_carousel_show_projects"),
            "img_max_height": "150" if is_coming_from_landing else None,
            **get_carousel_parameters(latest_publications.count(), content),
        },
    )


def get_carousel_parameters(number_of_results, content):
    number_of_latest_pub_to_display = PublicationsCustomization.get_int("publications_carousel_latest_display_count")
    interval_time_in_ms = PublicationsCustomization.get_int("publications_carousel_interval_time") * 1000
    tiles_per_item = PublicationsCustomization.get_int("publications_carousel_tiles_per_item", default=2)
    # we show them by pairs, so divide by 2 + 1 for the first item (leaderboard)
    items = 0
    if content in ["leaderboard", "mixed"]:
        items += 1
    if content in ["publications", "mixed"]:
        items += math.ceil(min(number_of_latest_pub_to_display, number_of_results) / tiles_per_item)
    refresh_time = items * interval_time_in_ms
    return {
        "interval_time": interval_time_in_ms,
        "refresh_time": refresh_time,
        "tiles_per_item": tiles_per_item,
        "item_grid_size": int(12 / tiles_per_item),
        "content": content,
    }


@login_required
@require_GET
def jumbotron(request, content: str = "mixed"):
    return render(request, "NEMO_publications/jumbotron.html", get_carousel_parameters(0, content))


def get_user_publication(user: User, doi: str) -> Optional[UserPublicationStatus]:
    return (
        UserPublicationStatus.objects.filter(user=user, status=UserPublicationStatus.Status.PENDING)
        .filter(doi__iexact=doi)
        .first()
    )


@login_required
@require_GET
def reject_user_publication(request, user_publication_status_id: int):
    user_pub = get_object_or_404(UserPublicationStatus, id=user_publication_status_id)
    if user_pub.status == UserPublicationStatus.Status.PENDING:
        user_pub.status = UserPublicationStatus.Status.REJECTED
        user_pub.save(update_fields=["status"])
    if UserPublicationStatus.objects.filter(user=user_pub.user, status=UserPublicationStatus.Status.PENDING).exists():
        return redirect("search_publication")
    return redirect("publications")



from django.apps import AppConfig
from django.conf import settings

from . import app_settings as defaults

# Set some app default settings
for name in dir(defaults):
    if name.isupper() and not hasattr(settings, name):
        setattr(settings, name, getattr(defaults, name))


class NemoPublicationsConfig(AppConfig):
    name = "NEMO_publications"
    verbose_name = "NEMO Publications"

    def ready(self):
        from django.utils.translation import gettext_lazy as _
        from NEMO.plugins.utils import (
            add_dynamic_notification_types,
            check_extra_dependencies,
        )
        from NEMO_publications.utils import PUBLICATION_NOTIFICATION

        check_extra_dependencies(self.name, ["NEMO", "NEMO-CE"])

        add_dynamic_notification_types(
            [(PUBLICATION_NOTIFICATION, _("Publications - notifies of suggested publications for users"))]
        )



# Generated by Django 4.2.23 on 2025-09-19 22:22

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ("NEMO_publications", "0005_userpublicationstatus"),
    ]

    operations = [
        migrations.AlterModelOptions(
            name="userpublicationstatus",
            options={"verbose_name_plural": "User publication statuses"},
        ),
    ]



from __future__ import annotations

from typing import List

from NEMO.models import User
from NEMO.utilities import export_format_datetime
from NEMO.views.api import ModelViewSet, datetime_filters, key_filters, manykey_filters, number_filters, string_filters
from django import forms
from django.db import transaction
from django.forms.models import model_to_dict
from django.http import QueryDict
from drf_excel.mixins import XLSXFileMixin
from rest_framework import status, viewsets
from rest_framework.exceptions import ValidationError
from rest_framework.response import Response

from NEMO_publications.forms import create_publication_metadata_post_form
from NEMO_publications.models import PublicationData, PublicationMetadata, UserPublicationStatus
from NEMO_publications.serializers import (
    PublicationDataSerializer,
    PublicationMetadataSerializer,
    PublicationSerializer,
    UserPublicationStatusSerializer,
)
from NEMO_publications.utils import fetch_publication_metadata_by_doi


class PublicationMetadataViewSet(ModelViewSet):
    filename = "publication_metadata"
    queryset = PublicationMetadata.objects.all()
    serializer_class = PublicationMetadataSerializer
    filterset_fields = {
        "id": key_filters,
        "creator": key_filters,
        "creation_time": datetime_filters,
        "doi": string_filters,
        "title": string_filters,
        "journal": string_filters,
        "year": number_filters,
        "month": number_filters,
        "status": number_filters,
        "decision_time": datetime_filters,
    }

    def create(self, request, *args, **kwargs):
        request_user = request.user
        serializer = self.get_serializer(data=request.data)
        datas = serializer.initial_data if getattr(serializer, "many", False) else [serializer.initial_data]
        created = [model_to_dict(self.create_publication_metadata(request_user, data)) for data in datas]
        headers = self.get_success_headers(serializer.initial_data)
        return Response(created, status=status.HTTP_201_CREATED, headers=headers)

    def update(self, request, *args, **kwargs):
        return Response({"detail": "Operation not supported."}, status=status.HTTP_404_NOT_FOUND)

    def create_publication_metadata(self, request_user: User, request_data: QueryDict):
        doi = request_data.get("doi")
        if not doi:
            raise ValidationError("DOI is required")
        if not PublicationMetadata.objects.filter(doi__iexact=doi).exists():
            # If metadata for DOI does not exist in DB, lookup in https://doi.org
            publication_metadata_search = fetch_publication_metadata_by_doi(doi)
            # Use user supplied data and override fields with results from DOI search
            metadata_data = request_data.copy()
            metadata_data.update(publication_metadata_search["metadata"])
            publication_metadata_form = create_publication_metadata_post_form(
                metadata_data, request_user, PublicationMetadata()
            )
            publication_metadata_form.instance.creator = request_user
            # If the form is not valid, display any previous errors
            if not publication_metadata_form.is_valid():
                if "error" in publication_metadata_search.keys():
                    publication_metadata_form.add_error(None, [publication_metadata_search["error"]])
                raise ValidationError(publication_metadata_form.errors)
            else:
                return publication_metadata_form.save()
        else:
            raise ValidationError({"error": ["Metadata already exists."]})


class PublicationDataViewSet(ModelViewSet):
    filename = "publication_data"
    queryset = PublicationData.objects.all()
    serializer_class = PublicationDataSerializer
    filterset_fields = {
        "id": key_filters,
        "creator": key_filters,
        "creation_time": datetime_filters,
        "metadata": key_filters,
        "authors": manykey_filters,
        "tools": manykey_filters,
        "projects": manykey_filters,
    }


class PublicationViewSet(XLSXFileMixin, viewsets.GenericViewSet):
    serializer_class = PublicationSerializer

    def check_permissions(self, request):
        if not request or not request.user or not request.user.is_active:
            self.permission_denied(request)

    def list(self, request, *args, **kwargs):
        publication_form = PublicationFilterForm(self.request.GET)
        if not publication_form.is_valid():
            return Response(status=status.HTTP_400_BAD_REQUEST, data=publication_form.errors)
        queryset = self.get_queryset()
        serializer = self.serializer_class(queryset, many=True)
        return Response(serializer.data)

    def get_queryset(self):
        publication_form = PublicationFilterForm(self.request.GET)
        publication_form.full_clean()
        data: List[Publication] = []

        publication_metadata_queryset = PublicationMetadata.objects.all()

        if publication_form.get_status():
            publication_metadata_queryset = publication_metadata_queryset.filter(status=publication_form.get_status())
        if publication_form.get_year():
            publication_metadata_queryset = publication_metadata_queryset.filter(year=publication_form.get_year())
        if publication_form.get_journal():
            publication_metadata_queryset = publication_metadata_queryset.filter(
                journal__iexact=publication_form.get_journal()
            )
        if publication_form.get_title():
            publication_metadata_queryset = publication_metadata_queryset.filter(
                title__iexact=publication_form.get_title()
            )
        if publication_form.get_doi():
            publication_metadata_queryset = publication_metadata_queryset.filter(doi__iexact=publication_form.get_doi())

        for publication in publication_metadata_queryset.all():
            publication_data_queryset = PublicationData.objects.filter(metadata=publication)
            if publication_form.get_tool_id():
                publication_data_queryset = publication_data_queryset.filter(tools__id=publication_form.get_tool_id())
            if publication_form.get_author_id():
                publication_data_queryset = publication_data_queryset.filter(
                    authors__id=publication_form.get_author_id()
                )
            if publication_form.get_project_id():
                publication_data_queryset = publication_data_queryset.filter(
                    projects__id=publication_form.get_project_id()
                )

            if publication_data_queryset.exists():
                data.append(Publication(publication, PublicationData.objects.filter(metadata=publication).all()))

        return data

    def get_filename(self, *args, **kwargs):
        return f"publications-{export_format_datetime()}.xlsx"


class Publication(object):
    def __init__(self, metadata: PublicationMetadata, entries: List[PublicationData]):
        self.doi = metadata.doi
        self.title = metadata.title
        self.journal = metadata.journal
        self.year = metadata.year
        self.status = metadata.status
        self.authors = set()
        self.projects = set()
        self.tools = set()
        for entry in entries:
            self.authors.update(entry.authors.all().values_list("id", flat=True))
            self.projects.update(entry.projects.all().values_list("id", flat=True))
            self.tools.update(entry.tools.all().values_list("id", flat=True))


class PublicationFilterForm(forms.Form):
    doi = forms.CharField(required=False)
    journal = forms.CharField(required=False)
    year = forms.CharField(required=False)
    title = forms.CharField(required=False)
    status = forms.IntegerField(required=False)
    author_id = forms.IntegerField(required=False)
    tool_id = forms.IntegerField(required=False)
    project_id = forms.IntegerField(required=False)

    def get_doi(self):
        return self.cleaned_data["doi"]

    def get_journal(self):
        return self.cleaned_data["journal"]

    def get_year(self):
        return self.cleaned_data["year"]

    def get_title(self):
        return self.cleaned_data["title"]

    def get_author_id(self):
        return self.cleaned_data["author_id"]

    def get_tool_id(self):
        return self.cleaned_data["tool_id"]

    def get_project_id(self):
        return self.cleaned_data["project_id"]

    def get_status(self):
        return self.cleaned_data["status"]


class UserPublicationStatusViewSet(ModelViewSet):
    filename = "user_publication_status"
    queryset = UserPublicationStatus.objects.all()
    serializer_class = UserPublicationStatusSerializer
    filterset_fields = {
        "id": key_filters,
        "user": key_filters,
        "doi": string_filters,
        "date_added": datetime_filters,
        "status": number_filters,
    }

    @transaction.atomic
    def create(self, request, *args, **kwargs):
        # We are overriding this to remove rows that would trigger the UniqueConstraint
        # Essentially, this allows uploading a list with duplicates, and it won't throw an error
        many = isinstance(request.data, list)
        if not many:
            return super().create(request, *args, **kwargs)
        else:
            serializer = self.get_serializer(data=request.data, many=many)
            safe_data = [data for data in request.data if not is_unique_constraint_error(serializer, data)]
            serializer = self.get_serializer(data=safe_data, many=many)
            serializer.is_valid(raise_exception=True)
            try:
                self.perform_create(serializer)
            except Exception as e:
                raise ValidationError({"error": str(e)})
            headers = self.get_success_headers(serializer.data)
            return Response(serializer.data, headers=headers)


def is_unique_constraint_error(serializer, data):
    try:
        serializer.run_child_validation(data)
        return False  # If no exception, the data is valid
    except ValidationError as exc:
        str_e = str(exc)
        return "user" in str_e and "doi" in str_e and "unique" in str_e



from NEMO.urls import router, sort_urls
from django.urls import path, re_path

from NEMO_publications import api
from NEMO_publications.views import publications

router.register(r"publications/publication_metadata", api.PublicationMetadataViewSet)
router.register(r"publications/publication_data", api.PublicationDataViewSet)
router.register(r"publications/publications", api.PublicationViewSet, basename="publication")
router.register(r"publications/user_publication_status", api.UserPublicationStatusViewSet)
router.registry.sort(key=sort_urls)

urlpatterns = [
    path("", publications.new_landing, name="landing"),
    path(
        "search_publication/",
        publications.search_publication_by_doi,
        name="search_publication",
    ),
    path(
        "edit_publication/<int:publication_data_id>/",
        publications.create_or_update_publication_data,
        name="edit_publication",
    ),
    path(
        "create_publication/<int:publication_metadata_id>/",
        publications.create_or_update_publication_data,
        name="create_publication",
    ),
    path(
        "save_publication_metadata/",
        publications.create_publication_metadata,
        name="save_publication_metadata",
    ),
    path(
        "delete_publication/<int:publication_data_id>/",
        publications.delete_publication,
        name="delete_publication",
    ),
    path(
        "publication_user_search/",
        publications.user_search,
        name="publication_user_search",
    ),
    path(
        "set_publication_status/<int:publication_metadata_id>/",
        publications.set_publication_status,
        name="set_publication_status",
    ),
    path("publications/", publications.get_publications, name="publications"),
    path(
        "reject_user_publication/<int:user_publication_status_id>/",
        publications.reject_user_publication,
        name="reject_user_publication",
    ),
    path("publications/leaderboard/", publications.get_publication_leaderboard, name="publications_leaderboard"),
    path(
        "publications/leaderboard/<int:year>/",
        publications.get_publication_leaderboard,
        name="publications_leaderboard",
    ),
    re_path(
        r"^publications/carousel_content/(?P<content>leaderboard|publications|mixed)/$",
        publications.carousel_content,
        name="publications_carousel_content",
    ),
    re_path(
        r"^publications/jumbotron/(?P<content>leaderboard|publications|mixed)/$",
        publications.jumbotron,
        name="publications_jumbotron",
    ),
    path("publications/jumbotron/", publications.jumbotron, name="publications_jumbotron"),
]



# Generated by Django 4.2.19 on 2025-05-23 13:21

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ("NEMO_publications", "0004_publicationmetadata_image_url"),
    ]

    operations = [
        migrations.CreateModel(
            name="UserPublicationStatus",
            fields=[
                ("id", models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                ("doi", models.CharField(max_length=255)),
                ("date_added", models.DateTimeField(auto_now_add=True)),
                (
                    "status",
                    models.PositiveIntegerField(choices=[(1, "Pending"), (2, "Added"), (3, "Rejected")], default=1),
                ),
                (
                    "user",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="publication_statuses",
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
            options={
                "unique_together": {("user", "doi")},
            },
        ),
    ]



from __future__ import annotations
from datetime import timedelta
from typing import TYPE_CHECKING

from NEMO.models import Notification
from NEMO.views.notifications import delete_notification
from django.contrib.contenttypes.models import ContentType
from django.utils import timezone

from NEMO_publications.customization import PublicationsCustomization
from NEMO_publications.utils import PUBLICATION_NOTIFICATION

if TYPE_CHECKING:
    from NEMO_publications.models import UserPublicationStatus


def manage_publication_notification(user_publication_status: UserPublicationStatus):
    from NEMO_publications.models import UserPublicationStatus

    if user_publication_status.status == UserPublicationStatus.Status.PENDING:
        expiration = timezone.now() + timedelta(
            days=PublicationsCustomization.get_int("publications_notification_expiration_days")
        )  # days for publication notifications to expire
        Notification.objects.get_or_create(
            user=user_publication_status.user,
            notification_type=PUBLICATION_NOTIFICATION,
            content_type=ContentType.objects.get_for_model(user_publication_status),
            object_id=user_publication_status.id,
            defaults={"expiration": expiration},
        )
    else:
        delete_publication_notification(user_publication_status)


def delete_publication_notification(user_publication_status: UserPublicationStatus):
    delete_notification(PUBLICATION_NOTIFICATION, user_publication_status.id, [user_publication_status.user])



from datetime import MAXYEAR, date

from NEMO.models import BaseModel, Project, Tool, User
from NEMO.views.constants import CHAR_FIELD_MAXIMUM_LENGTH
from django.core.validators import MaxValueValidator, MinValueValidator
from django.db import models
from django.db.models import Q
from django.db.models.signals import m2m_changed, post_delete, post_save
from django.dispatch import receiver

from NEMO_publications.notifications import delete_publication_notification, manage_publication_notification
from NEMO_publications.utils import sanitize_doi, parse_bibtex


class Months(models.IntegerChoices):
    JAN = 1, "JANUARY"
    FEB = 2, "FEBRUARY"
    MAR = 3, "MARCH"
    APR = 4, "APRIL"
    MAY = 5, "MAY"
    JUN = 6, "JUNE"
    JUL = 7, "JULY"
    AUG = 8, "AUGUST"
    SEP = 9, "SEPTEMBER"
    OCT = 10, "OCTOBER"
    NOV = 11, "NOVEMBER"
    DEC = 12, "DECEMBER"


class PublicationMetadataStatus(object):
    APPROVED = 1
    REJECTED = 2
    Choices = (
        (APPROVED, "Approved"),
        (REJECTED, "Rejected"),
    )


class PublicationMetadata(BaseModel):
    creator = models.ForeignKey(User, on_delete=models.CASCADE)
    creation_time = models.DateTimeField(
        auto_now_add=True,
        help_text="The date and time when the publication metadata was created.",
    )
    doi = models.CharField(
        max_length=CHAR_FIELD_MAXIMUM_LENGTH,
        help_text="Digital Object Identifier (DOI) of the publication.",
        unique=True,
    )
    title = models.CharField(
        max_length=CHAR_FIELD_MAXIMUM_LENGTH,
        help_text="Title of the publication.",
    )
    journal = models.CharField(
        max_length=CHAR_FIELD_MAXIMUM_LENGTH,
        help_text="Journal where the publication is published.",
    )
    bibtex = models.TextField(
        null=True,
        blank=True,
        help_text="Publication bibtex metadata.",
    )
    json_metadata = models.TextField(
        null=True,
        blank=True,
        help_text="Publication json metadata.",
    )
    year = models.PositiveIntegerField(
        help_text="Publication year.", validators=[MinValueValidator(1900), MaxValueValidator(MAXYEAR)]
    )
    month = models.PositiveIntegerField(null=True, blank=True, choices=Months.choices, help_text="Publication month.")
    image_url = models.URLField(null=True, blank=True, help_text="Publication image URL.")
    status = models.IntegerField(
        choices=PublicationMetadataStatus.Choices,
        null=True,
        blank=True,
        help_text="Publication metadata status.",
    )
    decision_time = models.DateTimeField(
        null=True,
        blank=True,
        help_text="The date and time when the publication metadata was approved or rejected.",
    )

    def get_doi_url(self):
        return "https://doi.org/" + self.doi

    def get_date(self) -> date:
        return date(year=self.year, month=self.month or 1, day=1)

    def get_authors(self, exclude=None):
        publication_data = self.publicationdata_set.all()
        if exclude:
            publication_data = publication_data.exclude(id__in=exclude)
        return set([author for pub in publication_data for author in pub.authors.all()])

    def get_projects(self, exclude=None):
        publication_data = self.publicationdata_set.all()
        if exclude:
            publication_data = publication_data.exclude(id__in=exclude)
        return set([project for pub in publication_data for project in pub.projects.all()])

    def get_tools(self, exclude=None):
        publication_data = self.publicationdata_set.all()
        if exclude:
            publication_data = publication_data.exclude(id__in=exclude)
        return set([tool for pub in publication_data for tool in pub.tools.all()])

    def get_related_data(self, exclude=None):
        return {
            "authors": self.get_authors(exclude),
            "tools": self.get_tools(exclude),
            "projects": self.get_projects(exclude),
        }

    def get_data_creators(self):
        data_creator_ids = self.publicationdata_set.values_list("creator").distinct()
        return User.objects.filter(id__in=data_creator_ids)

    def get_bibtex_fields(self):
        if self.bibtex:
            return parse_bibtex(self.bibtex)

    def get_bibtex_authors(self):
        parsed_bibtex = self.get_bibtex_fields()
        if parsed_bibtex:
            return parsed_bibtex.get("author") or parsed_bibtex.get("authors")

    def __str__(self):
        return self.title + " (" + self.doi + ")"

    def clean(self):
        self.doi = self.doi.upper()


class PublicationData(BaseModel):
    creator = models.ForeignKey(User, related_name="publication_data_creator", on_delete=models.CASCADE)
    creation_time = models.DateTimeField(
        auto_now_add=True,
        help_text="The date and time when the publication data was created.",
    )
    metadata = models.ForeignKey(
        PublicationMetadata,
        null=False,
        help_text="The metadata of the publication",
        on_delete=models.CASCADE,
    )
    authors = models.ManyToManyField(User, blank=True, help_text="Authors of the publication.")
    tools = models.ManyToManyField(Tool, blank=True, help_text="Tools involved in the publication.")
    projects = models.ManyToManyField(Project, blank=True, help_text="Projects involved in the publication.")

    def get_authors(self):
        return ", ".join([author.__str__() for author in self.authors.all()])

    def get_projects(self):
        return ", ".join([project.__str__() for project in self.projects.all()])

    def get_tools(self):
        return ", ".join([tool.__str__() for tool in self.tools.all()])


class PublicationDataM2MValidationError(Exception):
    def __init__(self, message):
        self.message = message


def validation_publication_data_m2m(data, creator):
    # Check that at least one of authors, projects, tools is specified
    if not data["authors"].exists() and not data["projects"].exists() and not data["tools"].exists():
        raise PublicationDataM2MValidationError("At least one author, project or tool is required.")
    # check that the creator is part of the projects added to the list
    if creator:
        if (
            data["projects"].exists()
            and not data["projects"].filter(Q(user=creator) | Q(manager_set=creator)).exists()
            and not creator.is_any_part_of_staff
        ):
            projects_str = ", ".join(
                [
                    project.__str__()
                    for project in data["projects"].exclude(Q(user=creator) | Q(manager_set=creator)).all()
                ]
            )
            raise PublicationDataM2MValidationError(
                {"projects": "You cannot add project(s): " + projects_str + " to the publication."}
            )


class UserPublicationStatus(BaseModel):
    class Status(models.IntegerChoices):
        PENDING = 1, "Pending"
        ADDED = 2, "Added"
        REJECTED = 3, "Rejected"

    user = models.ForeignKey(User, on_delete=models.CASCADE, related_name="publication_statuses")
    doi = models.CharField(max_length=CHAR_FIELD_MAXIMUM_LENGTH)
    date_added = models.DateTimeField(auto_now_add=True)
    status = models.PositiveIntegerField(choices=Status.choices, default=Status.PENDING)

    class Meta:
        unique_together = ("user", "doi")
        verbose_name_plural = "User publication statuses"

    def save(self, *args, **kwargs):
        self.doi = sanitize_doi(self.doi)
        super().save(*args, **kwargs)

    def __str__(self):
        return f"{self.user.username} - {self.doi} - {self.get_status_display()}"


@receiver(m2m_changed, sender=PublicationData.authors.through)
def handle_publication_authors_save(sender, instance: PublicationData, action, **kwargs):
    if action == "post_add":
        for pub_author in instance.authors.all():
            for user_publication in UserPublicationStatus.objects.filter(
                user=pub_author, doi=instance.metadata.doi, status=UserPublicationStatus.Status.PENDING
            ):
                user_publication.status = UserPublicationStatus.Status.ADDED
                user_publication.save(update_fields=["status"])


@receiver(post_save, sender=UserPublicationStatus)
def handle_user_publication_status_save(sender, instance, **kwargs):
    manage_publication_notification(instance)


@receiver(post_delete, sender=UserPublicationStatus)
def handle_user_publication_status_delete(sender, instance, **kwargs):
    delete_publication_notification(instance)



from django.core.exceptions import ValidationError
from django.forms import ModelForm

from NEMO_publications.models import (
    PublicationData,
    PublicationDataM2MValidationError,
    PublicationMetadata,
    UserPublicationStatus,
    validation_publication_data_m2m,
)
from NEMO_publications.utils import sanitize_doi


class PublicationDataForm(ModelForm):
    class Meta:
        model = PublicationData
        exclude = ["creation_time", "creator", "metadata"]

    def clean(self):
        cleaned_data = super().clean()
        try:
            validation_publication_data_m2m(cleaned_data, self.instance.creator)
            return cleaned_data
        except PublicationDataM2MValidationError as e:
            raise ValidationError(e.message)


class PublicationMetadataForm(ModelForm):
    class Meta:
        model = PublicationMetadata
        exclude = ["creation_time", "creator"]


class UserPublicationStatusForm(ModelForm):
    class Meta:
        model = UserPublicationStatus
        fields = "__all__"

    def clean_doi(self):
        return sanitize_doi(self.cleaned_data.get("doi"))


def create_publication_metadata_post_form(content, user, publication_metadata):
    return PublicationMetadataForm(
        content,
        instance=publication_metadata,
        initial={"creator": user},
    )


def create_publication_data_post_form(request, user, edit, publication_data):
    return PublicationDataForm(
        request.POST,
        instance=publication_data,
        initial={"creator": publication_data.creator if edit else user},
    )



# Generated by Django 3.2.22 on 2024-06-01 14:02

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("NEMO_publications", "0003_publicationmetadata_status_and_json"),
    ]

    operations = [
        migrations.AddField(
            model_name="publicationmetadata",
            name="image_url",
            field=models.URLField(blank=True, help_text="Publication image URL.", null=True),
        ),
    ]



import json
from argparse import Namespace
from typing import Optional
from urllib.parse import urlparse

import requests
from bibtexparser import bparser
from dateutil import parser
from django.conf import settings
from requests import Timeout

from NEMO_publications import app_settings

PUBLICATION_NOTIFICATION = "publications"


def get_publication_settings():
    pub_settings = app_settings.DEFAULT_PUBLICATION_SETTINGS
    pub_settings.update(settings.PUBLICATIONS_SETTINGS)
    return Namespace(**pub_settings)


def parse_bibtex(bibtex_string):
    bibtex_parsed = bparser.parse(bibtex_string)
    if len(bibtex_parsed.entries) != 1:
        raise Exception("multiple entries were found")
    else:
        return bibtex_parsed.entries[0]


# Fetch DOI metadata from https://doi.org
# key 'error' contains error message if an error was encountered
# key 'metadata' contains dict with metadata fields that were found
def fetch_publication_metadata_by_doi(doi):
    pub_settings = get_publication_settings()
    base_url = pub_settings.doi_search_url.format(doi)
    headers_options = pub_settings.headers
    result = {"metadata": {"doi": doi}}
    for headers in headers_options:
        result = {"metadata": {"doi": doi}}
        try:
            response = requests.get(base_url, headers=headers, timeout=pub_settings.timeout)
            if response.status_code == 200:
                bibtex_string = response.text.strip()
                try:
                    publication_parsed_metadata = parse_bibtex(bibtex_string)
                    result["metadata"] = {
                        "doi": publication_parsed_metadata.get("doi", doi),
                        "year": publication_parsed_metadata.get("year"),
                        "month": parse_month_from_metadata(publication_parsed_metadata.get("month")),
                        "journal": publication_parsed_metadata.get("journal"),
                        "title": publication_parsed_metadata["title"],
                        "bibtex": bibtex_string,
                        "json_metadata": fetch_json_metadata(doi),
                    }
                    return result
                except Exception as parse_error:
                    result["error"] = "Search returned invalid publication metadata, " + parse_error.__str__()
            elif response.status_code == 404:
                result["error"] = "Publication information was not found."
            else:
                result["error"] = "Search query has failed."
        except Timeout:
            result["error"] = (
                "The request timed out. Please try again or contact your administrator to increase the timeout"
            )
    return result


def fetch_json_metadata(doi):
    pub_settings = get_publication_settings()
    base_url = pub_settings.doi_search_url.format(doi)
    headers = {"Accept": "application/json"}
    try:
        response = requests.get(base_url, headers=headers, timeout=pub_settings.timeout)
        if response.status_code == 200:
            response_text = response.text.strip()
            parsed_json = json.loads(response_text)
            return json.dumps(parsed_json, indent=2)
        return None
    except Exception:
        return None


def parse_month_from_metadata(month_str) -> Optional[int]:
    if month_str:
        try:
            parsed_date = parser.parse(month_str)
            month_number = parsed_date.month
            return month_number
        except ValueError:
            pass
    return None


def sanitize_doi(doi: str):
    return urlparse(doi).path.lstrip("/")



# Potential tests to write
# Cannot overwrite metadata (using same DOI with different case)
# Searching DOI
# 	If DOI exists in DB use the one in DB
# 	If DOI doesn't exist in DB query doi.org
# 	If DOI doesn't exist in doi.org or incomplete metadata, return the metadata edit form
# Publication data
# 	Cannot add an empty entry
# 	Cannot add a project that the user is not a part of
# 	Cannot edit entry that does not belong to user
# 	Cannot delete entry that does not belong to user
# API
# Publication ViewSet
# 	Check that all params work together to return the right elements from the db
# Publication metadata (POST)
# 	If DOI exists in DB use the one in DB
# 	If DOI doesn't exist in DB query doi.org
# 	If DOI doesn't exist in doi.org or incomplete metadata, use the rest of the data in the query
# Publication metadata Update (POST)
# 	Check Operation unsupported



DEFAULT_PUBLICATION_SETTINGS = {
    "timeout": 15,
    "doi_search_url": "http://dx.doi.org/{}",
    "headers": [{"Accept": "application/x-bibtex"}, {"Accept": "text/bibliography; style=bibtex"}],
}

PUBLICATIONS_SETTINGS = {}



from typing import Dict

from NEMO.models import Project, Tool, User
from NEMO.serializers import ModelSerializer
from NEMO.templatetags.custom_tags_and_filters import app_installed
from NEMO.views.constants import CHAR_FIELD_MAXIMUM_LENGTH, CHAR_FIELD_MEDIUM_LENGTH
from rest_flex_fields.serializers import FlexFieldsSerializerMixin
from rest_framework import serializers
from rest_framework.fields import CharField, IntegerField, ListField

from NEMO_publications.models import (
    PublicationData,
    PublicationDataM2MValidationError,
    PublicationMetadata,
    UserPublicationStatus,
    validation_publication_data_m2m,
)
from NEMO_publications.utils import sanitize_doi


class PublicationMetadataSerializer(FlexFieldsSerializerMixin, ModelSerializer):
    class Meta:
        model = PublicationMetadata
        fields = "__all__"
        expandable_fields = {
            "creator": "NEMO.serializers.UserSerializer",
        }


class PublicationDataSerializer(FlexFieldsSerializerMixin, ModelSerializer):
    class Meta:
        model = PublicationData

        fields = "__all__"
        expandable_fields = {
            "metadata": "NEMO_publications.serializers.PublicationMetadataSerializer",
            "tools": ("NEMO.serializers.ToolSerializer", {"many": True}),
            "authors": ("NEMO.serializers.UserSerializer", {"many": True}),
            "projects": ("NEMO.serializers.ProjectSerializer", {"many": True}),
            "creator": "NEMO.serializers.UserSerializer",
        }

    def validate(self, data):
        try:
            m2m_fields = {
                "authors": User.objects.filter(id__in=[obj.id for obj in data["authors"]] if data["authors"] else []),
                "projects": Project.objects.filter(
                    id__in=[obj.id for obj in data["projects"]] if data["projects"] else []
                ),
                "tools": Tool.objects.filter(id__in=[obj.id for obj in data["tools"]] if data["tools"] else []),
            }
            validation_publication_data_m2m(m2m_fields, data["creator"])
            return data
        except PublicationDataM2MValidationError as e:
            raise serializers.ValidationError(e.message)


class PublicationSerializer(serializers.Serializer):
    doi = CharField(max_length=CHAR_FIELD_MAXIMUM_LENGTH, read_only=True)
    title = CharField(max_length=CHAR_FIELD_MAXIMUM_LENGTH, read_only=True)
    journal = CharField(max_length=CHAR_FIELD_MAXIMUM_LENGTH, read_only=True)
    year = IntegerField(read_only=True)
    status = IntegerField(read_only=True)
    authors = ListField()
    tools = ListField()
    projects = ListField()

    def update(self, instance, validated_data):
        pass

    def create(self, validated_data):
        pass

    class Meta:
        fields = "__all__"


# 2025-05-30 Remove when everyone is on NEMO-User-Details >= 1.11.1
USER_DETAILS_PLUGIN_FIELDS = ["employee_id", "orcid", "scopus_id", "researcher_id", "google_scholar_id"]


# This serializer allows sending either a username or a user id
# It now also checks for user details fields and attempts to find a user based on them if no user is present in the request data
class UserPublicationStatusSerializer(serializers.ModelSerializer):
    user = CharField(max_length=CHAR_FIELD_MEDIUM_LENGTH)

    def __init__(self, *args, **kwargs):
        if app_installed("NEMO_user_details"):
            from NEMO_user_details.admin import UserDetailsAdminForm

            # Create new dynamic CharFields for each of the relevant User Details fields.
            details_form = UserDetailsAdminForm()
            try:
                plugin_unique_field_names = details_form._meta.model.unique_identifier_field_names()
            except:
                plugin_unique_field_names = USER_DETAILS_PLUGIN_FIELDS
            for detail_field_name in plugin_unique_field_names:
                setattr(self, f"{detail_field_name}_details_field", details_form.fields.get(detail_field_name))
                if self.user_details_plugin_field(detail_field_name):
                    self.fields[detail_field_name] = CharField(
                        max_length=CHAR_FIELD_MEDIUM_LENGTH, required=False, write_only=True
                    )
        super().__init__(*args, **kwargs)

    class Meta:
        model = UserPublicationStatus
        fields = "__all__"

    def user_details_plugin_field(self, field_name):
        field = getattr(self, f"{field_name}_details_field", None)
        if field and not field.disabled:
            return field
        return None

    def to_representation(self, instance):
        representation = super().to_representation(instance)
        try:
            representation["user"] = instance.user.id
        except (ValueError, TypeError):
            representation["user"] = None
        return representation

    def validate_user(self, value):
        # Check if the input is a username or ID and fetch the user
        try:
            # Try to process as an ID
            if value.isdigit():
                return User.objects.get(id=value)
            else:
                # Otherwise assume it is a username
                return User.objects.get(username=value)
        except User.DoesNotExist:
            raise serializers.ValidationError("User not found. Provide a valid username or ID.")

    def to_internal_value(self, data):
        request_data = data.copy()
        if not request_data.get("user") and app_installed("NEMO_user_details"):
            user_details_field_values: Dict[str, str] = {}
            for detail_field_name in USER_DETAILS_PLUGIN_FIELDS:
                user_detail_field = self.user_details_plugin_field(detail_field_name)
                # get the value, but then remove it with pop so it doesn't get passed to UserPublicationStatus() (later) to create it'
                user_detail_field_value = request_data.get(detail_field_name, None)
                request_data.pop(detail_field_name, None)
                if user_detail_field and user_detail_field_value:
                    user_details_field_values[detail_field_name] = user_detail_field_value
            if user_details_field_values:
                # Attempt to find an existing user based on user details fields, if no user value is present
                user, field_name, value = get_user_by_detail_field(user_details_field_values)
                if user:
                    request_data["user"] = user.id
                else:
                    raise serializers.ValidationError(
                        {
                            field_name: [
                                f"Could not find a user with {self.user_details_plugin_field(field_name).label}: {value}."
                            ]
                        }
                    )

        return super().to_internal_value(request_data)

    def validate_doi(self, value):
        return sanitize_doi(value)


def get_user_by_detail_field(user_details_field_values: Dict[str, str]):
    field = None
    value = None

    from NEMO_user_details.models import UserDetails

    for field, value in user_details_field_values.items():
        try:
            return UserDetails.objects.get(**{field: field}).user, field, value
        except UserDetails.DoesNotExist:
            pass
    return None, field, value



from unittest import TestCase

from NEMO_publications.utils import sanitize_doi


class TestCleanDOI(TestCase):
    def test_clean_doi(self):
        test_cases = [
            ("https://doi.org/10.1000/j.journal.2023.01.001", "10.1000/j.journal.2023.01.001"),
            ("http://doi.org/10.1000/j.journal.2020.05.003", "10.1000/j.journal.2020.05.003"),
            ("https://dx.doi.org/10.1234/example.test.2021", "10.1234/example.test.2021"),
            ("http://dx.doi.org/10.5678/example.test.2022", "10.5678/example.test.2022"),
            ("10.9987/standalone.doi.example", "10.9987/standalone.doi.example"),
        ]

        for doi, expected in test_cases:
            with self.subTest(doi=doi):
                self.assertEqual(sanitize_doi(doi), expected)



# Generated by Django 3.2.22 on 2023-11-29 13:49

import django.core.validators
from NEMO_publications.utils import parse_bibtex, parse_month_from_metadata
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("NEMO_publications", "0001_initial"),
    ]

    def set_default_month_forward(apps, schema_editor):
        PublicationMetadata = apps.get_model("NEMO_publications", "PublicationMetadata")
        for metadata in PublicationMetadata.objects.all():
            if metadata.bibtex:
                publication_parsed_metadata = parse_bibtex(metadata.bibtex)
                month = parse_month_from_metadata(publication_parsed_metadata.get("month"))
                if month:
                    metadata.month = month
                    metadata.save()

    def set_default_month_reverse(apps, schema_editor):
        pass

    operations = [
        migrations.AddField(
            model_name="publicationmetadata",
            name="month",
            field=models.PositiveIntegerField(
                blank=True,
                choices=[
                    (1, "JANUARY"),
                    (2, "FEBRUARY"),
                    (3, "MARCH"),
                    (4, "APRIL"),
                    (5, "MAY"),
                    (6, "JUNE"),
                    (7, "JULY"),
                    (8, "AUGUST"),
                    (9, "SEPTEMBER"),
                    (10, "OCTOBER"),
                    (11, "NOVEMBER"),
                    (12, "DECEMBER"),
                ],
                help_text="Publication month.",
                null=True,
            ),
        ),
        migrations.RunPython(set_default_month_forward, set_default_month_reverse),
        migrations.AlterField(
            model_name="publicationmetadata",
            name="year",
            field=models.PositiveIntegerField(
                help_text="Publication year.",
                validators=[
                    django.core.validators.MinValueValidator(1900),
                    django.core.validators.MaxValueValidator(9999),
                ],
            ),
        ),
    ]



from NEMO.decorators import customization
from NEMO.views.customization import CustomizationBase


@customization(key="publications", title="Publications")
class PublicationsCustomization(CustomizationBase):
    variables = {
        "publications_enable_landing_carousel": "",
        "publications_carousel_latest_display_count": "10",
        "publications_carousel_latest_display_order": '["-creation_time"]',
        "publications_carousel_tiles_per_item": "2",
        "publications_carousel_interval_time": "5",
        "publications_carousel_show_date": "enabled",
        "publications_carousel_show_authors": "enabled",
        "publications_carousel_show_projects": "enabled",
        "publications_carousel_show_tools": "enabled",
        "publications_notification_expiration_days": "365",
        "publications_enable_landing_warning_pending_publications": "",
    }
